//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/MaskResultsPresenter.cpp
//! @brief     Implements class MaskResultsPresenter.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/MaskResultsPresenter.h"
#include "Device/Data/Datafield.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Mask/MaskStack.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MaskList.h"

namespace {

//! Constructs Datafield which contains original intensity data except masked areas,
//! and areas outside of ROI, where bin content is set to zero.

Datafield* createMaskPresentation(Data2DItem* data2DItem)
{
    // Requesting mask information
    std::unique_ptr<IShape2D> roi;
    Datafield* result = data2DItem->c_field()->clone();
    MaskStack detectorMask;
    const QVector<const MaskItem*> maskItems = data2DItem->maskContainerItem()->maskItems();
    for (auto maskIter = maskItems.rbegin(); maskIter != maskItems.rend(); maskIter++)
        if ((*maskIter)->isVisible()) {
            if (auto* roiItem = dynamic_cast<const RegionOfInterestItem*>((*maskIter)))
                roi = roiItem->createShape();
            else {
                std::unique_ptr<IShape2D> shape((*maskIter)->createShape());
                detectorMask.maskToStack(*shape, (*maskIter)->maskValue(), false);
            }
        }

    // ROI mask has to be the last one, it can not be "unmasked" by other shapes
    if (roi)
        detectorMask.maskToStack(*roi, true, false);

    if (!detectorMask.hasMasks())
        return nullptr;

    for (size_t i = 0; i < result->size(); ++i)
        if (detectorMask.isMasked(i, result->frame()))
            (*result)[i] = 0;

    return result;
}

} // namespace


MaskResultsPresenter::MaskResultsPresenter() = default;

MaskResultsPresenter::~MaskResultsPresenter() = default;

void MaskResultsPresenter::updateMaskResults(Data2DItem* data2DItem)
{
    m_data2DItem = data2DItem;
}

void MaskResultsPresenter::updatePresenter(bool pixelized)
{
    ASSERT(m_data2DItem);

    if (pixelized) {
        if (Datafield* maskedData = ::createMaskPresentation(m_data2DItem)) {
            // store data backup
            m_backup_data.reset(m_data2DItem->c_field()->clone());
            m_backup_interpolated = m_data2DItem->isInterpolated();

            m_data2DItem->setDatafield(*maskedData);
            m_data2DItem->setInterpolated(false);
        } else {
            m_backup_data.reset();
        }

    } else {
        if (m_backup_data) {
            m_data2DItem->setDatafield(*m_backup_data);
            m_data2DItem->setInterpolated(m_backup_interpolated);
        }
    }
}
