//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Mask/MaskList.cpp
//! @brief     Implements class MaskList.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Mask/MaskList.h"
#include "GUI/Model/Mask/MaskItems.h"
#include <QRegularExpression>

namespace {
namespace Tag {

const QString Mask("Mask");

} // namespace Tag
} // namespace

MaskList::MaskList() = default;

MaskList::~MaskList() = default;

QVector<MaskItem*> MaskList::modifiableMaskItems() const
{
    return m_maskItems.toModifiableQVector();
}

QVector<const MaskItem*> MaskList::maskItems() const
{
    return m_maskItems.toQVector();
}

void MaskList::insertMask(int row, MaskItem* maskItem)
{
    // takes owning of maskItem!
    m_maskItems.insert_at(row, maskItem);
}

void MaskList::addMaskItem(MaskItem* maskItem)
{
    // takes owning of maskItem!
    m_maskItems.push_back(maskItem);
}

void MaskList::moveMask(int from_row, int to_row)
{
    m_maskItems.move(from_row, to_row);
}

void MaskList::removeMaskAt(int row)
{
    m_maskItems.delete_at(row);
}

RegionOfInterestItem* MaskList::regionOfInterestItem() const
{
    for (const auto& maskSel : m_maskItems)
        if (auto* roi = dynamic_cast<RegionOfInterestItem*>(maskSel.currentItem()))
            return roi;

    return nullptr;
}

void MaskList::clear()
{
    m_maskItems.clear();
}

bool MaskList::isEmpty() const
{
    return m_maskItems.empty();
}

int MaskList::size() const
{
    return m_maskItems.size();
}

MaskItem* MaskList::at(const int idx)
{
    return m_maskItems.at(idx).currentItem();
}

int MaskList::indexOfItem(const MaskItem* maskItem) const
{
    return m_maskItems.index_of(maskItem);
}

void MaskList::updateMaskNames()
{
    const auto reg = QRegularExpression("[0-9]");

    QMap<QString, int> numMasksByType;
    for (const auto& m_maskItem : m_maskItems) {

        QString name = m_maskItem.currentItem()->maskName();
        name.remove(reg);

        int numMasks = 1;
        if (numMasksByType.contains(name)) {
            numMasks = numMasksByType.value(name) + 1;
            numMasksByType.remove(name);
        }
        numMasksByType.insert(name, numMasks);
        name += QString::number(numMasks);

        m_maskItem.currentItem()->setMaskName(name);
    }
}

void MaskList::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    for (const auto& sel : m_maskItems) {
        w->writeStartElement(Tag::Mask);
        sel.writeTo(w);
        w->writeEndElement();
    }
}

void MaskList::readFrom(QXmlStreamReader* r, MessageService*)
{
    clear();

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        if (tag == Tag::Mask) {
            addMaskItem(nullptr);
            m_maskItems.back().readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}
