//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/DatafilesModel.cpp
//! @brief     Implements class DatafilesModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/DatafilesModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {
namespace Tag {

const QString RealData("RealData");
const QString SelectedIndex("SelectedIndex");
const QString SelectedRank("SelectedRank");

} // namespace Tag
} // namespace

DatafilesModel::DatafilesModel()
{
    setObjectName("DatafilesModel");
}

DatafilesModel::~DatafilesModel() = default;

QVector<DataItem*> DatafilesModel::dataItems() const
{
    QVector<DataItem*> result;
    for (auto* realData : realItems())
        if (auto* data2DItem = realData->dataItem())
            result.push_back(data2DItem);

    return result;
}

void DatafilesModel::clear()
{
    for (auto* realItem : realItems())
        removeDatafileItem(realItem);
}

void DatafilesModel::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // real items
    for (auto* realItem : realItems()) {
        w->writeStartElement(Tag::RealData);
        XML::writeAttribute(w, XML::Attrib::name, realItem->realItemName());
        realItem->writeTo(w);
        w->writeEndElement();
    }

    // selected index
    w->writeStartElement(Tag::SelectedIndex);
    XML::writeAttribute(w, XML::Attrib::value, m_selectedIndex);
    w->writeEndElement();

    // selected rank
    w->writeStartElement(Tag::SelectedRank);
    XML::writeAttribute(w, XML::Attrib::value, m_selectedRank);
    w->writeEndElement();
}

void DatafilesModel::readFrom(QXmlStreamReader* r)
{
    clear();

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // real item
        if (tag == Tag::RealData) {
            pushNewItem()->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // selected index
        } else if (tag == Tag::SelectedIndex) {
            XML::readAttribute(r, XML::Attrib::value, &m_selectedIndex);
            XML::gotoEndElementOfTag(r, tag);

            // selected rank
        } else if (tag == Tag::SelectedRank) {
            XML::readAttribute(r, XML::Attrib::value, &m_selectedRank);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }

    if (r->hasError())
        throw std::runtime_error(r->errorString().toLatin1());
}

void DatafilesModel::writeDatafiles(const QString& projectDir)
{
    for (const auto* realItem : realItems())
        realItem->writeDatafiles(projectDir);

    dataFilesCleaner.cleanOldFiles(projectDir, dataItems());
}

void DatafilesModel::readDatafiles(const QString& projectDir, MessageService* messageService)
{
    for (auto* realItem : realItems())
        realItem->readDatafiles(projectDir, messageService);

    dataFilesCleaner.recollectDataNames(dataItems());
}

DatafileItem* DatafilesModel::pushNewItem()
{
    m_realItems.emplace_back(new DatafileItem);
    return m_realItems.back().get();
}

DatafileItem* DatafilesModel::insertDataItem(int rank)
{
    DatafileItem* r = pushNewItem();
    r->initItemWithRank(rank);
    return r;
}

QVector<DatafileItem*> DatafilesModel::realItems() const
{
    QVector<DatafileItem*> items(m_realItems.size());
    for (int i = 0; i < items.size(); i++)
        items[i] = m_realItems[i].get();
    return items;
}


QVector<DatafileItem*> DatafilesModel::realItems(int rank) const
{
    return rank == 1 ? realItems1D() : realItems2D();
}

QVector<DatafileItem*> DatafilesModel::realItems1D() const
{
    QVector<DatafileItem*> vec_1d;

    for (auto* item : realItems())
        if (item->rank() == 1)
            vec_1d.push_back(item);

    return vec_1d;
}

QVector<DatafileItem*> DatafilesModel::realItems2D() const
{
    QVector<DatafileItem*> vec_2d;

    for (auto* item : realItems())
        if (item->rank() == 2)
            vec_2d.push_back(item);

    return vec_2d;
}

void DatafilesModel::removeDatafileItem(DatafileItem* realItem)
{
    ASSERT(realItem);
    int index = realItems().indexOf(realItem);
    m_realItems.erase(m_realItems.begin() + index);
}

QStringList DatafilesModel::realItemNames() const
{
    QStringList result;
    for (auto* item : realItems())
        result.append(item->realItemName());
    return result;
}
