//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Data/DatafilesSelector.cpp
//! @brief     Implements class DatafilesSelector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Data/DatafilesSelector.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Model/DatafilesTree.h"
#include "GUI/View/Data/DatafileEditor.h"
#include "GUI/View/Data/Legacy1dDialog.h"
#include "GUI/View/Info/MessageBox.h"
#include "GUI/View/Item/ItemViewOverlayButtons.h"
#include "GUI/View/Manager/ProjectManager.h"
#include "GUI/View/Widget/ApplicationSettings.h"
#include <QFileDialog>
#include <QMenu>
#include <QSplitter>
#include <QVBoxLayout>

namespace {

// The following two functions are templated on T in {IO::Filetype1D, IO::Filetype2D}.

template <class T>
QString join_filterkeys(std::vector<std::pair<const QString, T>> _nomap, const QString& _separator)
{
    QString result;
    for (const auto& it : _nomap) {
        if (!result.isEmpty())
            result += _separator;
        result += it.first;
    }
    return result;
}

template <class T>
T filterkey2type(std::vector<std::pair<const QString, T>> _nomap, const QString& _key)
{
    for (const auto& it : _nomap)
        if (it.first == _key)
            return it.second;
    ASSERT_NEVER;
}

} // namespace


DatafilesSelector::DatafilesSelector(QWidget* parent, ProjectDocument* document)
    : QWidget(parent)
    , m_import1dDataAction(new QAction(this))
    , m_import2dDataAction(new QAction(this))
    , m_renameDataAction(new QAction(this))
    , m_removeDataAction(new QAction(this))
    , m_treeView(new QTreeView(this))
    , m_treeModel(new DatafilesTree(this, document->realModel()))
    , m_editor(new DatafileEditor(this, document))
    , m_document(document)
{
    setMinimumSize(250, 600);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
    setWindowTitle("DatafilesSelector");

    m_import2dDataAction->setText("Import 2D data");
    m_import2dDataAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_import2dDataAction->setIconText("2D");
    m_import2dDataAction->setToolTip("Import 2D data");
    connect(m_import2dDataAction, &QAction::triggered, [this] { importData2D(); });

    m_import1dDataAction->setText("Import 1D data");
    m_import1dDataAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_import1dDataAction->setIconText("1D");
    m_import1dDataAction->setToolTip("Import 1D data");
    connect(m_import1dDataAction, &QAction::triggered, [this] { importData1D(); });

    m_renameDataAction->setText("Rename");
    m_renameDataAction->setIcon(QIcon()); // #baTODO: Icon needed?
    m_renameDataAction->setIconText("Rename");
    m_renameDataAction->setToolTip("Rename data");
    connect(m_renameDataAction, &QAction::triggered, this, &DatafilesSelector::renameCurrentItem);

    m_removeDataAction->setText("Remove");
    m_removeDataAction->setIcon(QIcon(":/images/delete.svg"));
    m_removeDataAction->setIconText("Remove");
    m_removeDataAction->setToolTip("Remove selected data");
    connect(m_removeDataAction, &QAction::triggered, this, &DatafilesSelector::removeCurrentItem);

    // m_treeView->setAttribute(Qt::WA_MacShowFocusRect, false); // TODO/Mac: needed?
    m_treeView->setItemsExpandable(false);
    m_treeView->setRootIsDecorated(false);
    m_treeView->setHeaderHidden(true);
    m_treeView->setContextMenuPolicy(Qt::CustomContextMenu);
    m_treeView->setModel(m_treeModel);

    auto* splitter = new QSplitter;
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_treeView);
    splitter->addWidget(m_editor);
    splitter->setChildrenCollapsible(true);

    auto* mainLayout = new QVBoxLayout;
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->setSpacing(0);
    mainLayout->addWidget(splitter);

    setLayout(mainLayout);

    connect(m_editor, &DatafileEditor::instrumentUpdated, this,
            &DatafilesSelector::selectionChanged);

    connect(m_treeView, &QTreeView::customContextMenuRequested, this,
            &DatafilesSelector::onContextMenuRequest);

    connect(m_treeModel, &QAbstractItemModel::modelReset, [this] { m_treeView->expandAll(); });

    connect(m_treeModel, &QAbstractItemModel::rowsInserted, [this] { m_treeView->expandAll(); });

    connect(m_treeView->selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &DatafilesSelector::onSelectionChanged, Qt::UniqueConnection);

    m_treeView->expandAll();
    updateActionEnabling();

    restoreSelection();

    ItemViewOverlayButtons::install(
        m_treeView, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
}

QSize DatafilesSelector::sizeHint() const
{
    return {200, 400};
}

QSize DatafilesSelector::minimumSizeHint() const
{
    return {128, 200};
}

DatafileItem* DatafilesSelector::currentItem()
{
    return m_treeModel->itemForIndex(currentIndex());
}

void DatafilesSelector::setCurrentItem(DatafileItem* item)
{
    m_treeView->selectionModel()->clearSelection();
    QModelIndex index = m_treeModel->indexForItem(item);
    if (index.isValid())
        m_treeView->selectionModel()->setCurrentIndex(index, QItemSelectionModel::SelectCurrent);
}

void DatafilesSelector::restoreSelection()
{
    int lastIndex = m_document->realModel()->selectedIndex();
    int lastRank = m_document->realModel()->selectedRank();
    QModelIndex lastUsedIndex =
        m_treeModel->index(lastIndex, 0, m_treeModel->indexOfHeadline(lastRank));

    if (lastUsedIndex.isValid())
        m_treeView->selectionModel()->setCurrentIndex(lastUsedIndex,
                                                      QItemSelectionModel::SelectCurrent);
    else
        setCurrentItem(m_treeModel->topMostItem());

    if (currentItem())
        emit selectionChanged(currentItem());
}

QModelIndex DatafilesSelector::currentIndex()
{
    if (!m_treeView->selectionModel()->isSelected(m_treeView->selectionModel()->currentIndex()))
        return {};
    return m_treeView->selectionModel()->currentIndex();
}

QList<QAction*> DatafilesSelector::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (m_treeModel->isHeadline(index)) {
        if (asHover)
            return {};
        return (index == m_treeModel->indexOfHeadline(1)) ? QList<QAction*>({m_import1dDataAction})
                                                          : QList<QAction*>({m_import2dDataAction});
    }

    // -- index belongs to item
    if (!asHover)
        return {};

    auto* item = m_treeModel->itemForIndex(index);
    if (item == nullptr)
        return {};

    auto* removeAction = new QAction(this);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this data");
    connect(removeAction, &QAction::triggered, [this, item] {
        m_treeModel->removeDataItem(item);
        gProjectDocument.value()->setModified();
    });

    return {removeAction};
}

void DatafilesSelector::onSelectionChanged()
{
    updateActionEnabling();

    m_document->realModel()->setSelectedIndex(currentIndex().row());
    m_document->realModel()->setSelectedRank(currentIndex().parent().row() + 1);

    m_editor->setDatafileItem(currentItem());
    emit selectionChanged(currentItem());
}

void DatafilesSelector::onContextMenuRequest(const QPoint& point)
{
    auto* realItemAtPoint = m_treeModel->itemForIndex(m_treeView->indexAt(point));
    updateActionEnabling(realItemAtPoint);

    QMenu menu;
    menu.setToolTipsVisible(true);

    if (realItemAtPoint != nullptr) {
        menu.addAction(m_renameDataAction);
        menu.addAction(m_removeDataAction);
        menu.addSeparator();
    }

    menu.addAction(m_import2dDataAction);
    menu.addAction(m_import1dDataAction);
    menu.exec(m_treeView->mapToGlobal(point));
}

void DatafilesSelector::importData1D()
{
    static const std::vector<std::pair<const QString, IO::Filetype1D>> filters1D{
        {"Legacy table / interactive configuration (*.csv *.dat *.tab *.txt *.dat.gz)", IO::csv1D},
        {"Motofit (*.mft)", IO::mft},
        {"BornAgain (*.int *.int.gz)", IO::bornagain1D},
        {"all (*.*)", IO::unknown1D}};

    static const QString filters = ::join_filterkeys(filters1D, ";;");

    QString selectedFilter = ProjectManager::instance()->recentlyUsedImportFilter1D();
    const QString dirname = ProjectManager::instance()->userImportDir();

    const QStringList fnames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    if (fnames.isEmpty())
        return;

    ProjectManager::instance()->setImportDirFromFilePath(fnames[0]);
    ProjectManager::instance()->setRecentlyUsedImportFilter1D(selectedFilter);

    const IO::Filetype1D global_ftype = ::filterkey2type(filters1D, selectedFilter);

    for (const QString& fname : fnames) {
        DatafileItem* realItem = nullptr;
        const ImportSettings1D* settings = nullptr;

        IO::Filetype1D ftype = global_ftype;
        if (ftype == IO::unknown1D)
            ftype = IO::filename2type1D(fname.toStdString());
        if (ftype == IO::csv1D) {
            Legacy1dDialog dialog(this, fname);
            const auto result = dialog.exec();
            if (result != QDialog::Accepted)
                return;
            settings = &Legacy1dDialog::Msettings;
        }

        try {
            std::unique_ptr<Datafield> df(IO::readData1D(fname.toStdString(), ftype, settings));
            if (!df)
                continue;
            if (df->rank() != 1)
                throw std::runtime_error("File does not contain a 1d data set");
            realItem = m_treeModel->injectDataItem(1);
            realItem->setDatafileItemName(QFileInfo(fname).baseName());
            realItem->setTheDatafield(*df);
            setCurrentItem(realItem);

            realItem->updateFileName();
        } catch (std::exception& ex) {
            m_treeModel->removeDataItem(realItem);

            const QString message = QString("Cannot read file %1:\n\n%2")
                                        .arg(fname)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::Message::warning(this, "File import", message);
        }
    }

    gProjectDocument.value()->setModified();
}

void DatafilesSelector::importData2D()
{
    static const std::vector<std::pair<const QString, IO::Filetype2D>> filters2D{
        {"all (*.*)", IO::unknown2D},
        {"TIFF (*.tif *.tiff *.tif.gz)", IO::tiff},
        {"Nicos/SANSDRaw (*.001)", IO::nicos2D},
        {"BornAgain (*.int *.int.gz)", IO::bornagain2D},
        {"CSV (*.csv *.dat *.tab *.txt *.dat.gz)", IO::csv2D},
        {"REFSANS (*.csv)", IO::refsans2D}};

    static const QString filters = ::join_filterkeys(filters2D, ";;");

    QString selectedFilter = ProjectManager::instance()->recentlyUsedImportFilter2D();
    const QString dirname = ProjectManager::instance()->userImportDir();

    const QStringList fnames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    if (fnames.isEmpty())
        return;

    ProjectManager::instance()->setImportDirFromFilePath(fnames[0]);
    ProjectManager::instance()->setRecentlyUsedImportFilter2D(selectedFilter);

    const IO::Filetype2D global_ftype = ::filterkey2type(filters2D, selectedFilter);

    for (const QString& fname : fnames) {
        IO::Filetype2D ftype = global_ftype;
        if (ftype == IO::unknown2D)
            ftype = IO::filename2type2D(fname.toStdString());

        DatafileItem* realItem = nullptr;
        try {
            std::unique_ptr<Datafield> df(IO::readData2D(fname.toStdString(), ftype));
            if (!df)
                continue;
            if (df->rank() != 2)
                throw std::runtime_error("File does not contain a 2d data set");
            realItem = m_treeModel->injectDataItem(2);
            realItem->setDatafileItemName(QFileInfo(fname).baseName());
            realItem->setTheDatafield(*df);
            setCurrentItem(realItem);
        } catch (std::exception& ex) {
            m_treeModel->removeDataItem(realItem);

            const QString message = QString("Cannot read file %1\n\n%2")
                                        .arg(fname)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::Message::warning(this, "File import", message);
        }
    }

    gProjectDocument.value()->setModified();
}

void DatafilesSelector::renameCurrentItem()
{
    if (currentIndex().isValid()) {
        // First delete any existing index widget. This is necessary if overlay buttons are used,
        // because these buttons use the indexWidget() functionality, which itself uses the same
        // internals as an item editor
        m_treeView->setIndexWidget(currentIndex(), nullptr);
        m_treeView->edit(currentIndex());
        gProjectDocument.value()->setModified();
    }
}

void DatafilesSelector::removeCurrentItem()
{
    m_treeModel->removeDataItem(currentItem());
    gProjectDocument.value()->setModified();
}

void DatafilesSelector::updateActionEnabling()
{
    updateActionEnabling(currentItem());
}

void DatafilesSelector::updateActionEnabling(const DatafileItem* item) const
{
    m_removeDataAction->setEnabled(item != nullptr);
    m_renameDataAction->setEnabled(item != nullptr);
}
