//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/Plot2DWithProjectionFrame.cpp
//! @brief     Defines class Plot2DWithProjectionFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/Plot2DWithProjectionFrame.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Mask/ProjectionList.h"
#include "GUI/View/Canvas/ProjectionsEditorCanvas.h"
#include "GUI/View/Canvas/ProjectionsWidget.h"
#include "GUI/View/Setup/ProjectionActions.h"
#include "GUI/View/Setup/ProjectionsToolpanel.h"
#include "GUI/View/Setup/Scale2DEditor.h"
#include <QBoxLayout>
#include <QSplitter>

Plot2DWithProjectionFrame::Plot2DWithProjectionFrame(QWidget* parent)
    : DataAccessWidget(parent)
    , m_editorActions(new ProjectionActions(this))
    , m_toolbar(new ProjectionsToolpanel(m_editorActions))
    , m_projectionsCanvas(new ProjectionsEditorCanvas)
    , m_projectionsWidget(new ProjectionsWidget)
    , m_scaleEditor(new Scale2DEditor(this))
{
    auto* layout = new QHBoxLayout;
    setLayout(layout);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);

    auto* splitter = new QSplitter;
    layout->addWidget(splitter);
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_projectionsCanvas);
    splitter->addWidget(m_projectionsWidget);

    layout->addWidget(m_scaleEditor);
    m_scaleEditor->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
    m_scaleEditor->setHidden(true);

    m_toolbar->setOrientation(Qt::Vertical);
    m_toolbar->setStyleSheet("QToolBar{border-left:1px solid rgb(180,180,180);}");
    layout->addWidget(m_toolbar);

    setup_connections();
}

void Plot2DWithProjectionFrame::setJobOrDatafileItem(QObject* item)
{
    setBaseItem(item);
    m_scaleEditor->setJobOrDatafileItem(item);
    updateProjsFrame();
}

void Plot2DWithProjectionFrame::updateProjsFrame()
{
    ASSERT(currentData2DItem());

    // The following line must come before 'm_projectionsCanvas->setContext()'!
    // It prevents problem with switching between datasets in projection mode.
    m_projectionsWidget->disconnectItem();

    m_projectionsCanvas->updateProjsCanvas(currentData2DItem());
    m_projectionsWidget->setData2DItem(currentData2DItem());

    m_editorActions->updateProjsActions(currentData2DItem());
}

void Plot2DWithProjectionFrame::showEvent(QShowEvent*)
{
    if (currentData2DItem())
        updateProjsFrame();
}

void Plot2DWithProjectionFrame::hideEvent(QHideEvent*)
{
    m_projectionsCanvas->resetProjsCanvas();
    m_projectionsWidget->disconnectItem();
    m_projectionsWidget->setData2DItem(nullptr);
}

QList<QAction*> Plot2DWithProjectionFrame::actionList()
{
    return m_editorActions->topToolboxActions();
}

void Plot2DWithProjectionFrame::setup_connections()
{
    // tool panel request is propagated from editorActions to this MaskEditor
    connect(m_editorActions, &ProjectionActions::resetViewRequest, m_projectionsCanvas,
            &ProjectionsEditorCanvas::onResetViewRequest, Qt::UniqueConnection);

    // tool panel request is propagated from editorActions to this MaskEditor
    connect(m_editorActions, &ProjectionActions::propertyPanelRequest,
            [this] { m_scaleEditor->setHidden(!m_scaleEditor->isHidden()); });

    // selection/drawing activity is propagated from Toolbar to graphics scene
    connect(m_toolbar, &ProjectionsToolpanel::activityChanged, m_projectionsCanvas,
            &ProjectionsEditorCanvas::onActivityChanged, Qt::UniqueConnection);

    // selection/drawing activity is propagated from Toolbar to Projections Widget
    connect(m_toolbar, &ProjectionsToolpanel::activityChanged, m_projectionsWidget,
            &ProjectionsWidget::onActivityChanged, Qt::UniqueConnection);

    // click on projections tab is propagated to tool bar
    connect(m_projectionsWidget, &ProjectionsWidget::changeActivityRequest, m_toolbar,
            &ProjectionsToolpanel::onProjectionTabChange, Qt::UniqueConnection);

    // Delete request is propagated from canvas to actions
    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::deleteSelectedRequest, m_editorActions,
            &ProjectionActions::onDeleteAction, Qt::UniqueConnection);

    // moving projection automatically switches projections tab
    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::changeProjectionsTabRequest,
            m_projectionsWidget, &ProjectionsWidget::onActivityChanged, Qt::UniqueConnection);

    // space bar push (request for zoom mode) is propagated from graphics view to Toolbar
    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::changeActivityRequest, m_toolbar,
            &ProjectionsToolpanel::onChangeActivityRequest, Qt::UniqueConnection);

    // ColorMap margins changed, canvas -> projection widget
    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::marginsChanged, m_projectionsWidget,
            &ProjectionsWidget::onMarginsChanged, Qt::UniqueConnection);

    m_toolbar->onChangeActivityRequest(MaskFlags::HORIZONTAL_LINE_MODE);
}
