//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Project/SimulationView.h
//! @brief     Defines class SimulationView.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_PROJECT_SIMULATIONVIEW_H
#define BORNAGAIN_GUI_VIEW_PROJECT_SIMULATIONVIEW_H

#include <QCheckBox>
#include <QComboBox>
#include <QLabel>
#include <QPushButton>
#include <QRadioButton>
#include <QSpinBox>

class DatafileItem;
class InstrumentItem;
class MainWindow;
class ProjectDocument;
class SampleItem;
class SimulationOptionsItem;

//! Widget to define a simulation.
//! Contains:
//! * Elements to select instrument, sample and real data.
//! * Elements to set options like number of threads.
//! * Buttons to run simulation or to export as a Python script
class SimulationView : public QWidget {
    Q_OBJECT
public:
    SimulationView(QWidget* parent, ProjectDocument* document);

    void simulate();
    void exportPythonScript();

private:
    void showEvent(QShowEvent*) override;
    void hideEvent(QHideEvent*) override;

    //! Read the entries in the UI and modify document.
    void updateStateFromUI();

    //! Read the entries in the UI and store in OptionsItem.
    void readOptionsFromUI();

    //! Write entries in OptionsItem to UI
    void writeOptionsToUI();

    //! Update enabling of elements depending on other elements.
    void updateEnabling();

    //! Returns selected sample taking into account that there might be several
    //! samples with the same name.
    const SampleItem* selectedSampleItem() const;

    //! Returns selected InstrumentItem taking into account that there might be several
    //! instruments with the same name.
    const InstrumentItem* selectedInstrumentItem() const;

    //! Returns selected real data item taking into account that there might be several
    //! items with the same name.
    const DatafileItem* selectedDatafileItem() const;

    //! Updates selection combo with string list while preserving previous selection.
    //! If allowNone == true, additional "None" item will be added to the combo.
    void updateSelection(QComboBox* comboBox, QStringList itemList, int currentIndex,
                         bool allowNone = false);

    //! Checks whether selection is correct for running a simulation.
    //! Returns empty string if valid, otherwise the error text.
    QString validateSimulationSetup(bool validateRealData) const;

    ProjectDocument* m_document;

    QLabel* m_instrumentLabel;
    QComboBox* m_instrumentCombo;
    QComboBox* m_sampleCombo;
    QComboBox* m_realDataCombo;

    QRadioButton* m_analyticalRadio;
    QRadioButton* m_monteCarloRadio;
    QSpinBox* m_numberOfMonteCarloPoints;

    QRadioButton* m_ambientLayerRadio;
    QRadioButton* m_averageLayerRadio;

    QCheckBox* m_includeSpecularCheck;

    QRadioButton* m_runPolicyImmediatelyRadio;
    QRadioButton* m_runPolicyBackgroundRadio;
    QComboBox* m_numberOfThreadsCombo;

    QPushButton* m_simulateButton;
    QPushButton* m_exportToPyScriptButton;

    // Convenience method for easier access
    SimulationOptionsItem* optionsItem() const;
};

#endif // BORNAGAIN_GUI_VIEW_PROJECT_SIMULATIONVIEW_H
