//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Scene/MaskGraphicsScene.h
//! @brief     Defines class MaskGraphicsScene.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_SCENE_MASKGRAPHICSSCENE_H
#define BORNAGAIN_GUI_VIEW_SCENE_MASKGRAPHICSSCENE_H

#include "GUI/View/Scene/MaskFlags.h"
#include <QGraphicsScene>
#include <QItemSelection>
#include <QItemSelectionModel>
#include <memory>

class ColorMap;
class Data2DItem;
class IOverlay;
class LineItem;
class MaskEditorAction;
class MaskGraphicsProxy;
class MaskItem;
class MaskeditorListmodel;
class OverlayItem;
class PolygonOverlay;

//! Associated with MaskGraphicsViews, holds ColorMap and masks.

class MaskGraphicsScene : public QGraphicsScene {
    Q_OBJECT
public:
    MaskGraphicsScene();
    ~MaskGraphicsScene() override;

    void associateItems(Data2DItem* data_item, MaskeditorListmodel* mask_list_model,
                        QItemSelectionModel* selectionModel);
    void dissociateItems();
    void updateSize(const QSize& newSize);
    void cancelCurrentDrawing();

    ColorMap* colorMap() { return m_plot; }

signals:
    void itemContextMenuRequest(const QPoint& point);
    void lineItemProcessed();
    void lineItemMoved(LineItem* sender);
    void lineItemDeleted(const LineItem* sender);

public slots:
    void onActivityChanged(MaskFlags::MaskMode mask_mode);
    void onMaskValueChanged(MaskFlags::MaskValue mask_value);
    void onRowsAboutToBeRemoved(const QModelIndex&, int first, int last);

private slots:
    void onSessionSelectionChanged(const QItemSelection& /* selected */,
                                   const QItemSelection& /* deselected */);

private:
    void mousePressEvent(QGraphicsSceneMouseEvent* event) override;
    void mouseMoveEvent(QGraphicsSceneMouseEvent* event) override;
    void mouseReleaseEvent(QGraphicsSceneMouseEvent* event) override;
    void drawForeground(QPainter* painter, const QRectF& rect) override;
    void contextMenuEvent(QGraphicsSceneContextMenuEvent* event) override;

    void resetScene();
    void updateScene();
    void onSceneSelectionChanged();
    void updateProxyWidget();
    void updateOverlays();
    IOverlay* registerOverlay(OverlayItem* item);
    void removeOverlay(OverlayItem* item);

    void setDrawingInProgress(bool value);

    void makeViewAtMousePosSelected(QGraphicsSceneMouseEvent* event);

    void processRectangleOrEllipseItem(QGraphicsSceneMouseEvent* event);
    void processPolygonItem(QGraphicsSceneMouseEvent* event);
    void processLineItem(QGraphicsSceneMouseEvent* event);
    void processVerticalLineItem(const QPointF& pos);
    void processHorizontalLineItem(const QPointF& pos);
    void processFullframeItem(QGraphicsSceneMouseEvent* event);

    bool isValidMouseClick(QGraphicsSceneMouseEvent* event) const;
    bool isValidForRectangleShapeDrawing(QGraphicsSceneMouseEvent* event) const;
    bool isValidForPolygonDrawing(QGraphicsSceneMouseEvent* event) const;
    bool isValidForLineDrawing(QGraphicsSceneMouseEvent* event) const;
    bool isValidForMaskAllDrawing(QGraphicsSceneMouseEvent* event) const;

    PolygonOverlay* currentPolygon() const;

    ColorMap* m_plot;
    MaskeditorListmodel* m_mask_list_model = nullptr;
    QItemSelectionModel* m_selection_model = nullptr;
    QMap<const OverlayItem*, IOverlay*> m_mask2overlay;
    MaskGraphicsProxy* m_proxy = nullptr;
    bool m_block_selection = false;
    bool m_mouse_is_pressed = false;
    Data2DItem* m_data_item = nullptr;
    MaskItem* m_active_mask = nullptr;
    QPointF m_mouse_position;
    bool m_drawing_in_progress;
    MaskFlags::MaskValue m_mask_value;
    MaskFlags::MaskMode m_mask_mode;
};

#endif // BORNAGAIN_GUI_VIEW_SCENE_MASKGRAPHICSSCENE_H
